@@ -10,7 +10,6 @@ from gitlab.mixins import (
     UpdateMixin,
 )
 
-
 __all__ = [
     "ProjectApproval",
     "ProjectApprovalManager",
@@ -73,7 +72,8 @@ class ProjectApprovalRuleManager(
     _obj_cls = ProjectApprovalRule
     _from_parent_attrs = {"project_id": "id"}
     _create_attrs = RequiredOptional(
-        required=("name", "approvals_required"), optional=("user_ids", "group_ids")
+        required=("name", "approvals_required"),
+        optional=("user_ids", "group_ids", "protected_branch_ids"),
     )
 
 
